"""A package with very important objects in structural engineering domain.

The package is the most important of *pycivil* and contains a part of the
codebase in structural engineering domain. Starting from it, we can build
tools like post-processors (for Strand7, Midas etc.), checkers (for
standard codes like Eurocodes, Italian Codes).

"""