"""A package with tools built around pycivil and other python libraries

The package contains tools like post-processors (for Strand7, Midas etc.),
checkers (for standard codes like Eurocodes, Italian Codes) and wrappers
around important python packages like vtk or gmsh.
"""